/*
								+----------------------------------+
								|                                  |
								| *** DigiLib OpenGL extension *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|         DigiLib_Ext.cpp          |
								|                                  |
								+----------------------------------+
*/

#include <vector>
#include "../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <gl/gl.h>
#include <gl/glext.h>
#include <algorithm>
#include <stdlib.h>
#include <string.h>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLDrv.h"
#include "../berLame_src/OpenGLState.h"
#include "../berLame_src/Texture.h"
#include "../berLame_src/RenderBuffer2.h"
#include "../berLame_src/Shader2.h"
#include "../berLame_src/ShaderSys.h"
#include "../berLame_src/VertexBufferObject.h"
#include "DigiLib_Ext.h"

/*
 *	struct TImageStructInfo
 *		- structure, containing associated data for ImageStruct
 */
struct TImageStructInfo {
	ImageStruct *p_image_struct;
	bool b_current_on_server;
	CGLTexture_2D *p_texture;

	bool b_pixel_buffer_object;
	CPixelBufferObject *p_pbo; // contains image, mapped to system memory

	void Delete()
	{
		if(p_texture) {
			delete p_texture;
			p_texture = 0;
		}
		if(p_pbo) {
			if(p_pbo->b_Mapped())
				p_pbo->UnMap();
			delete p_pbo;
			p_pbo = 0;
		}
		if(p_image_struct) {
			if(b_pixel_buffer_object)
				p_image_struct->Raster = NULL; // raster is no longer valid
			free(p_image_struct); // allocated using malloc() in p_Create_GL_Image()
			p_image_struct = 0;
		}
	}
};

static bool b_pixel_buffer_supported = false;
static bool b_fence_supported = false;
static bool b_fbo_supported = false;
static bool b_gl_images_supported = false;
// opengl capabilities

static std::vector<TImageStructInfo> imagestruct_list;
// list of textures, bound to imagestructs

static CGLState *p_state = 0;
// opengl state cache

static bool b_auto_uploads = true;
static bool b_auto_downloads = true;
// automatic uploads / downloads

/*
 *								--- function objects for imagestruct_list ---
 */

class CFindImageStructInfo {
protected:
	const ImageStruct *m_p_image_struct;

public:
	__inline CFindImageStructInfo(const ImageStruct *p_image_struct)
		:m_p_image_struct(p_image_struct)
	{}

	__inline bool operator ()(const TImageStructInfo &r_info)
	{
		return r_info.p_image_struct == m_p_image_struct;
	}
};

class CDeleteImageStructInfo {
public:
	__inline void operator ()(TImageStructInfo &r_info)
	{
		r_info.Delete();
	}
};

/*
 *								--- ~function objects for imagestruct_list ---
 */

/*
 *								--- win32 OpenGL ---
 */

#ifdef WIN32
static HWND h_wnd;
static CGLDriver *p_gl_driver = 0;

unsigned __int32 *p_DecompressRLEImage(const unsigned __int32 *p_image, int n_size)
{
	unsigned __int32 *p_buffer;
	if(!(p_buffer = new unsigned __int32[n_size]))
		return 0;

	const unsigned __int32 *p_end = p_buffer + n_size;
	for(unsigned __int32 *p_dest = p_buffer; p_dest != p_end;) {
		bool b_compress = (*p_image >> 31) != 0;
		int n_run_length = *p_image & ((1u << 31) - 1);

		++ p_image; // skip run length

		if(b_compress) {
			unsigned __int32 n_value = *p_image;
			++ p_image;
			// get color

			for(unsigned __int32 *p_end2 = p_dest + n_run_length; p_dest != p_end2; ++ p_dest)
				*p_dest = n_value;
			// fill
		} else {
			for(unsigned __int32 *p_end2 = p_dest + n_run_length;
			   p_dest != p_end2; ++ p_dest, ++ p_image)
				*p_dest = *p_image;
			// copy data
		}
	}
	// decompress

	return p_buffer;
}

bool Create_Window(const char *p_s_name, int n_width, int n_height)
{
	WNDCLASSEX t_wnd_class;
	t_wnd_class.cbSize = sizeof(WNDCLASSEX);
	t_wnd_class.style = CS_HREDRAW | CS_VREDRAW;
	t_wnd_class.lpfnWndProc = DefWindowProc;
	t_wnd_class.cbClsExtra = 0;
	t_wnd_class.cbWndExtra = 0;
	t_wnd_class.hInstance = GetModuleHandle(NULL);
	t_wnd_class.hIcon = LoadIcon(GetModuleHandle(NULL), IDI_APPLICATION);
	t_wnd_class.hCursor = LoadCursor(GetModuleHandle(NULL), IDC_ARROW);
	t_wnd_class.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	t_wnd_class.lpszMenuName = 0;
	t_wnd_class.lpszClassName = p_s_name;
	t_wnd_class.hIconSm = LoadIcon(GetModuleHandle(NULL), IDI_APPLICATION);
	if(!RegisterClassEx(&t_wnd_class))
		return false;
	// create window class

	RECT t_rect;
	t_rect.left = t_rect.top = 0;
	t_rect.right = n_width;
	t_rect.bottom = n_height;
	if(!AdjustWindowRect(&t_rect, WS_BORDER | WS_CAPTION, false))
		return false;
	// calculate window size to contain client area with desired resolution

	if(!(h_wnd = CreateWindow(p_s_name, p_s_name,
	   WS_BORDER | WS_CAPTION, CW_USEDEFAULT, CW_USEDEFAULT,
	   t_rect.right - t_rect.left, t_rect.bottom - t_rect.top,
	   NULL, NULL, GetModuleHandle(NULL), NULL)))
		return false;
	// create window

	ShowWindow(h_wnd, SW_SHOW);
	UpdateWindow(h_wnd);
	// show & update window

	return true;
}

#endif //WIN32

/*
 *								--- ~win32 OpenGL ---
 */

bool Init_OpenGL() // vraci true, pokud se podarilo. pod linuxem pouziva glut, pod windows pouzije COpenGLDriver z berLame
{
	if(p_state)
		return true;

#ifdef WIN32
	if(!Create_Window("DigiLib_Ext", 150, 57))
		return false;
	// create window

	if(!(p_gl_driver = new CGLDriver()))
		return false;
	if(!p_gl_driver->Init(h_wnd, 150, 57, 24, 0)) {
		delete p_gl_driver;
		p_gl_driver = 0;
		return false;
	}
	// init open-gl

	if(!(p_state = new CGLState(wglGetCurrentDC(), wglGetCurrentContext()))) {
		delete p_gl_driver;
		p_gl_driver = 0;
		return false;
	}
	// create state guard
#else //WIN32
	// todo -  linux opengl (init glut)
#endif //WIN32

	b_fbo_supported = CGLRenderBuffer_FBO::b_Supported(); // gets function pointers as well
	CGLExtensionHandler::n_GetGL12FuncPointers();
	CGLExtensionHandler::n_GetGL13FuncPointers();
	CGLExtensionHandler::n_GetMultitextureFuncPointers(); // opengl1.3 as well
	CGLExtensionHandler::n_GetGL14FuncPointers();
	CGLExtensionHandler::n_GetGL15FuncPointers();
	CGLExtensionHandler::n_GetGL20FuncPointers();
	b_pixel_buffer_supported = CPixelBufferObject::b_Supported(); // gets function pointers as well
	b_gl_images_supported = CPixelBufferObject::b_Supported();
	b_fence_supported = CGLExtensionHandler::b_SupportedExtension("GL_NV_fence") &&
		!CGLExtensionHandler::n_GetFenceNVFuncPointers();
	// get function pointers

	/*unsigned char *p_gl_logo = p_DecompressRLEImage(p_opengl_logo_2_0_sm_jpg,
		n_opengl_logo_2_0_sm_jpg_width * n_opengl_logo_2_0_sm_jpg_height);
	if(p_gl_logo) {
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		CGLTexture_2D opengl_tex(p_state, n_opengl_logo_2_0_sm_jpg_width,
			n_opengl_logo_2_0_sm_jpg_height, GL_LUMINANCE8, false, 0, GL_TEXTURE_2D,
			GL_LUMINANCE, GL_UNSIGNED_BYTE, p_gl_logo);
		delete[] p_gl_logo;*/
	// grayscale logo

	#include "OpenGL20Logo.h"
	unsigned __int32 *p_gl_logo = p_DecompressRLEImage(p_opengl_logo_2_0_sm_jpg,
		n_opengl_logo_2_0_sm_jpg_width * n_opengl_logo_2_0_sm_jpg_height);
	if(p_gl_logo) {
		/*glDrawPixels(n_opengl_logo_2_0_sm_jpg_width, n_opengl_logo_2_0_sm_jpg_height,
			GL_BGRA, GL_UNSIGNED_BYTE, p_gl_logo);*/ // upside-down
		CGLTexture_2D opengl_tex(p_state, n_opengl_logo_2_0_sm_jpg_width,
			n_opengl_logo_2_0_sm_jpg_height, GL_RGBA8, false, 0, GL_TEXTURE_2D,
			GL_BGRA, GL_UNSIGNED_BYTE, p_gl_logo);
		delete[] p_gl_logo;

		p_state->DisableDepthTest();
		p_state->DisableCullFace();
		p_state->Color3f(1, 1, 1);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		opengl_tex.Bind_Enable(p_state);
		glBegin(GL_QUADS);
		glTexCoord2f(0, 1);
		glVertex2f(-1, -1);
		glTexCoord2f(0, 0);
		glVertex2f(-1,  1);
		glTexCoord2f(1, 0);
		glVertex2f( 1,  1);
		glTexCoord2f(1, 1);
		glVertex2f( 1, -1);
		glEnd();

		p_gl_driver->Blit();
	}
	// load OpenGL logo

	return true;
}

bool Free_OpenGL_Objects() // odalokuje vsechny shadery, textury, buffery
{
	std::for_each(imagestruct_list.begin(), imagestruct_list.end(), CDeleteImageStructInfo());
	imagestruct_list.clear();
	// delete associated images

	return true;
}

bool Shutdown_OpenGL() // ukonci OpenGL
{
	if(!p_state)
		return true;

	bool b_result = Free_OpenGL_Objects();

	delete p_state;
	p_state = 0;

#ifdef WIN32
	b_result &= p_gl_driver->Shutdown();
	delete p_gl_driver;
	// shutdown open-gl

	DestroyWindow(h_wnd);
	// destroy window
#else
	// todo - linux opengl (shutdown glut)
#endif

	return b_result;
}

bool b_NPOTTexturesSupported() // vraci true, pokud OpenGL podporuje textury, jejichz rozmer neni mocnina dvou
{
	return CGLExtensionHandler::b_SupportedExtension("GL_ARB_texture_non_power_of_two");
}

bool b_FloatTexturesSupported() // vraci true, pokud OpenGL podporuje textury s float (IEEE754 32-bit) pixely; je k dispozici i rychlejsi 16 bit format, jez ma ovsem mensi rozsah (+-65030) a pri presunech z a do pameti se musi pouzivat konverze, takze se mu vyhybame
{
	return CGLExtensionHandler::b_SupportedExtension("GL_ARB_texture_float");
}

bool b_FramebufferObjectSupported() // vraci true, pokud OpenGL podporuje framebuffer objekty. bez nich nejde renderovat primo do textury, coz prakticky znemoznuje jakekoliv univerzalnejsi zpracovani obrazu
{
	return b_fbo_supported/*CGLRenderBuffer_FBO::b_Supported()*/;
}

int n_MaxTextureSize()
{
	GLint n_size;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &n_size);
	return n_size;
}

int n_PixelSize(int n_pixel_type)
{
	switch(n_pixel_type) {
	case Image8:
	case Image8Linear:
		return sizeof(unsigned char);
	case Image16:
	case Image16Linear:
		return sizeof(short);
	case Image32:
	case Image32Linear:
		return sizeof(long);
	case ImageFloat:
	case ImageFloatLinear:
		return sizeof(float);
	case ImageRGB:
	case ImageRGBLinear:
		return sizeof(unsigned long);
	case ImageComplex:
	case ImageComplexLinear:
		return 2 * sizeof(float);
	default:
		return 0;
	}
}

bool Get_GL_PixelType(int n_pixel_type, int &r_n_fmt, int &r_n_data_type, int &r_n_internal_fmt)
{
	r_n_fmt = GL_LUMINANCE; // 1 channel only, converts to red, 0, 0 when displayed
	r_n_data_type = GL_UNSIGNED_BYTE;

	switch(n_pixel_type) {
	case ImageRGB:
	case ImageRGBLinear:
		r_n_fmt = GL_BGRA;
		r_n_internal_fmt = GL_BGRA;
		break;
	case Image8:
	case Image8Linear:
		//r_n_data_type = GL_UNSIGNED_BYTE;
		r_n_internal_fmt = GL_LUMINANCE8;
		break;
	case Image16:
	case Image16Linear:
		r_n_data_type = GL_SHORT;
		r_n_internal_fmt = GL_LUMINANCE16;
		break;
	case Image32:
	case Image32Linear:
		r_n_data_type = GL_INT;
		r_n_internal_fmt = GL_LUMINANCE16; // beware rounding !!!
		break;
	case ImageFloat:
	case ImageFloatLinear:
		r_n_data_type = GL_FLOAT;
		r_n_internal_fmt = GL_LUMINANCE32F_ARB;
		break;
	case ImageComplex:
	case ImageComplexLinear:
		r_n_fmt = GL_LUMINANCE_ALPHA; // two - channel luminance-alpha image
		r_n_data_type = GL_FLOAT;
		r_n_internal_fmt = GL_LUMINANCE_ALPHA32F_ARB;
		break;
	default:
		return false;
	}

	return true;
}

bool b_FastTransfer(const ImageStruct *p_image) // vraci true, pokud se obrazek bude z/do OpenGL kopirovat rychle (tzn. jednim blokovym presunem)
{
	int n_pixel_size = n_PixelSize(p_image->PixelType);

	if(abs(p_image->XOffset) > 4 || abs(p_image->XOffset) < 1)
		return false;
	// pixels are interleaved

	if(p_image->YOffset != n_pixel_size * p_image->XSize &&
	   p_image->YOffset != -n_pixel_size * p_image->XSize)
	    return false;
	// scanlines are interleaved

	return true;
}

CGLTexture_2D *p_Create_ImageStruct_Texture(const ImageStruct *p_image)
{
	int n_format;
	int n_data_type;
	int n_internal_format;

	if(!Get_GL_PixelType(p_image->PixelType, n_format, n_data_type, n_internal_format))
		return 0;

	void *p_raster = p_image->Raster;
	int n_y_offset;
	if(p_image->YOffset >= 0)
		n_y_offset = p_image->YOffset;
	else {
		n_y_offset = -p_image->YOffset;
		p_raster = (void*)((int)p_raster - n_y_offset * p_image->YSize);
	}
	int n_x_offset;
	if(p_image->XOffset >= 0)
		n_x_offset = p_image->XOffset;
	else {
		n_x_offset = -p_image->XOffset;
		p_raster = (void*)((int)p_raster - n_x_offset * p_image->XSize);
	}
	// find upper-left corner of image

	if(n_y_offset != n_x_offset * p_image->XSize) {
		_ASSERTE(n_y_offset > n_x_offset * p_image->XSize); // y-offset must be grater than 
		glPixelStorei(GL_UNPACK_ROW_LENGTH, n_y_offset);
	} else
		glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
	// set proper scanline interleaving

	if(n_x_offset >= 1 && n_x_offset <= 4)
		glPixelStorei(GL_UNPACK_ALIGNMENT, n_x_offset);
	else {
		CGLTexture_2D *p_texture;
		if(!(p_texture = new CGLTexture_2D(p_state, p_image->XSize, p_image->YSize,
		   n_internal_format, false, 0, GL_TEXTURE_2D)))
			return 0;
		if(!p_texture->b_Status()) {
			delete p_texture;
			return 0;
		}
		// create empty texture

		p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
		p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
		// clamp to edge

		int n_x = 0;
		for(void *p_end = (void*)((int)p_raster + p_image->XSize * n_x_offset);
		   p_raster != p_end; p_raster = (void*)((int)p_raster + n_x_offset), ++ n_x) {
			glTexSubImage2D(GL_TEXTURE_2D, 0, n_x, 0, 1, p_image->YSize,
				n_format, n_data_type, p_raster);
		}
		// copy image as vertical scanlines (slow!)

		return p_texture;
	}
	// set proper pixel alignment

	CGLTexture_2D *p_texture;
	if(!(p_texture = new CGLTexture_2D(p_state, p_image->XSize, p_image->YSize,
	   n_internal_format, false, 0, GL_TEXTURE_2D, n_format, n_data_type, p_raster)))
		return 0;
	if(!p_texture->b_Status()) {
		delete p_texture;
		return 0;
	}
	// create texture

	p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
	p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
	// clamp to edge

	return p_texture;
}

bool b_HasTexture(const ImageStruct *p_image) // vraci true, pokud obrazek ma asociovanou nejakou texturu
{
	return std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
		CFindImageStructInfo(p_image)) != imagestruct_list.end();
}

bool Upload_ImageStruct(ImageStruct *p_image) // nahraje obrazek do graficke karty
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image))) == imagestruct_list.end()) {
	    imagestruct_list.reserve(imagestruct_list.size() + 1);
		if(imagestruct_list.capacity() < imagestruct_list.size() + 1)
			return false;
		// no more place in the list

		TImageStructInfo t_temp;
		t_temp.b_current_on_server = true;
		t_temp.b_pixel_buffer_object = false;
		t_temp.p_image_struct = p_image;
		t_temp.p_pbo = 0;
		if(!(t_temp.p_texture = p_Create_ImageStruct_Texture(p_image)))
			return false;
		// create texture

		imagestruct_list.push_back(t_temp);
		p_image_info = imagestruct_list.end() - 1;
		// put it into list, update iterator
	}
	// make sure there is associated texture

	// todo - upload

	return true;
}

bool Download_ImageStruct(ImageStruct *p_image) // stahne obrazek z graficke karty
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image))) != imagestruct_list.end()) {
		// todo - download

	    return true;
	}

	return false;
	// has no associated texture
}

bool b_AsyncSupported() // vrati true, pokud jsou podporovany asynchronni prenosy
{
	return b_pixel_buffer_supported;//CVertexBufferObject::b_Supported() && CPixelBufferObject::b_Supported();
}

bool b_FenceSupported() // vrati true, pokud je podporovan NV_FENCE (zjistovani dokonceni operaci, ati stale nema ekvivalent)
{
	return b_fence_supported;//CGLExtensionHandler::b_SupportedExtension("GL_NV_fence");
}

bool Upload_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id); // asynchronne nahraje obrazek do graficke karty; do p_fence_id nahraje id plotu
bool Download_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id); // asynchronne stahne obrazek do graficke karty; do p_fence_id nahraje id plotu

bool b_FenceFinished(unsigned int n_fence_id) // vrati true, pokud je operace za niz nasledoval plot n_fence_id dokoncena; jinak false
{
	if(!b_fence_supported) {
		glFlush();
		glFinish();
		return true;
	}
	return glTestFenceNV(n_fence_id) != 0;
}

bool WaitFence(unsigned int n_fence_id) // ceka dokud plot n_fence_id neskonci
{
	if(!b_fence_supported) {
		glFlush();
		glFinish();
		return true;
	}
	if(!glIsFenceNV(n_fence_id)) {
		glFlush();
		glFinish();
		return false;
	}
	glFinishFenceNV(n_fence_id);
	return true;
}

void Enable_AutoUploads() // povoli automaticke uploady (default; obrazek se zkopiruje ze sve pameti pokazde kdyz se s nim provadi nejaka operace)
{
	b_auto_uploads = true;
}

void Enable_AutoDownloads() // povoli automaticke downloady (default; obrazek se zkopiruje z karty pokazde kdyz se do nej provadi nejaka operace)
{
	b_auto_downloads = true;
}

void Disable_AutoUploads()
{
	b_auto_uploads = false;
}

void Disable_AutoDownloads() // zakaze automaticke uploady / downloady
{
	b_auto_downloads = false;
}

bool Set_DirtyFlag(const ImageStruct *p_image, bool b_current_on_server) // nastavi priznak modifikace obrazu. b_current_on_server rika zda je obrazek aktualni na graficke karte (true) nebo v systemove pameti (false); prvni funkce jez s obrazkem bude operovat potom muze provest potrebnou kopii a vynulovat priznak (tzn. bud nejaka operace nad obrazkem nebo [Up/Down]load[_Async]_ImageStruct())
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image))) != imagestruct_list.end()) {
		(*p_image_info).b_current_on_server = b_current_on_server; // set attrib
	    return true;
	}

	return b_current_on_server == false;
	// in case we wanted to invalidate server-side copy, we haven't failed
	// (has no associated texture)
}

bool b_GL_ImagesSupported() // vrati true pokud jsou podporovany obrazky, lezici v graficke pameti mapovane do systemove ram
{
	return b_gl_images_supported;
}

bool b_IsGL_Image(const ImageStruct *p_image) // vrati true, pokud je dany imagestruct takovym obrazkem
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image))) != imagestruct_list.end())
	    return (*p_image_info).b_pixel_buffer_object;
	return false;
}

bool b_GL_Image_RelocTest() // vrati true, pokud se adresy dat obrazku mezi jednotlivymi operacemi meni (pokusi se to vyzkouset, negativni vysledek nemusi odpovidat pravde)
{
	// todo
	return true;
}

ImageStruct *p_Create_GL_Image(int n_width, int n_height, short n_format) // vytvori mapovany obrazek o danych rozmerech a formatu (vzdycky je to linearni blok pixelu; nepodporuje tedy formaty jez nekonci _linear)
{
	int n_internal_format, n_fmt, n_data;
	if(!Get_GL_PixelType(n_format, n_fmt, n_data, n_internal_format))
		return 0;
	// get internal format

	imagestruct_list.reserve(imagestruct_list.size() + 1);
	if(imagestruct_list.capacity() < imagestruct_list.size() + 1)
		return 0;
	// no more place in the list

	TImageStructInfo t_temp;
	t_temp.b_current_on_server = false;
	t_temp.b_pixel_buffer_object = true;
	if(!(t_temp.p_texture = new CGLTexture_2D(p_state, n_width, n_height, n_internal_format, false)))
		return 0;
	t_temp.p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
	t_temp.p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
	// create texture, set clamp to edge

	if(!(t_temp.p_pbo = new CPixelBufferObject)) {
		delete t_temp.p_texture;
		return 0;
	}
	if(!t_temp.p_pbo->Alloc(n_PixelSize(n_format) * n_width * n_height)) {
		delete t_temp.p_texture;
		delete t_temp.p_pbo;
		return 0;
	}
	// create pixel-buffer object to be associated with texture // todo - don't remember how to bind

	ImageStruct *p_image_struct;
	if((p_image_struct = (ImageStruct*)malloc(sizeof(ImageStruct))) == NULL) {
		delete t_temp.p_texture;
		delete t_temp.p_pbo;
		return 0;
	}
	p_image_struct->PixelType = n_format;
	p_image_struct->XOffset = n_PixelSize(n_format);
	p_image_struct->XSize = n_width;
	p_image_struct->YOffset = n_width * n_PixelSize(n_format);
	p_image_struct->YSize = n_height;
	p_image_struct->Raster = (unsigned char*)t_temp.p_pbo->p_Map(GL_READ_WRITE_ARB);
	p_image_struct->Optional.Regular.Free = Destroy_GL_Image;
	// fill imagestruct

	t_temp.p_image_struct = p_image_struct;

	imagestruct_list.push_back(t_temp);
	// put it into list

	return p_image_struct;
}

void Destroy_GL_Image(void *p_gl_image) // zrusi mapovany obrazek (sem ukazuje pointer Free v ImageStruct)
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo((ImageStruct*)p_gl_image))) != imagestruct_list.end()) {
	    (*p_image_info).Delete();
		imagestruct_list.erase(p_image_info, p_image_info + 1);
	}
	// find and destroy TImageStructInfo and OpenGL objects

	free(p_gl_image);
	// free imagestruct
}

/*
 *		-end-of-file-
 */
